@@ -1,17 +1,17 @@
-import datetime
-
 from django.db import models
 from django.conf import settings
 from django.db.models import Q
 from django.core.cache import cache
+from django.core.exceptions import ValidationError
 
+from django.utils import timezone
 from django.utils.translation import ugettext_lazy as _
 
 from friendship.signals import friendship_request_created, \
-        friendship_request_rejected, friendship_request_canceled, \
-        friendship_request_viewed, friendship_request_accepted, \
-        friendship_removed, follower_created, following_created, follower_removed,\
-        following_removed
+    friendship_request_rejected, friendship_request_canceled, \
+    friendship_request_viewed, friendship_request_accepted, \
+    friendship_removed, follower_created, following_created, follower_removed,\
+    following_removed
 
 AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')
 
@@ -20,6 +20,7 @@ CACHE_TYPES = {
     'followers': 'fo-%d',
     'following': 'fl-%d',
     'requests': 'fr-%d',
+    'sent_requests': 'sfr-%d',
     'unread_requests': 'fru-%d',
     'unread_request_count': 'fruc-%d',
     'read_requests': 'frr-%d',
@@ -36,7 +37,7 @@ BUST_CACHES = {
         'unread_request_count',
         'read_requests',
         'rejected_requests',
-        ],
+    ],
 }
 
 
@@ -63,7 +64,7 @@ class FriendshipRequest(models.Model):
 
     message = models.TextField(_('Message'), blank=True)
 
-    created = models.DateTimeField(default=datetime.datetime.now)
+    created = models.DateTimeField(default=timezone.now)
     rejected = models.DateTimeField(blank=True, null=True)
     viewed = models.DateTimeField(blank=True, null=True)
 
@@ -78,18 +79,20 @@ class FriendshipRequest(models.Model):
     def accept(self):
         """ Accept this friendship request """
         relation1 = Friend.objects.create(
-                        from_user=self.from_user,
-                        to_user=self.to_user
-                    )
+            from_user=self.from_user,
+            to_user=self.to_user
+        )
+
         relation2 = Friend.objects.create(
-                        from_user=self.to_user,
-                        to_user=self.from_user
-                    )
+            from_user=self.to_user,
+            to_user=self.from_user
+        )
+
         friendship_request_accepted.send(
-                sender=self,
-                from_user=self.from_user,
-                to_user=self.to_user
-            )
+            sender=self,
+            from_user=self.from_user,
+            to_user=self.to_user
+        )
 
         self.delete()
         bust_cache('requests', self.to_user.pk)
@@ -99,7 +102,7 @@ class FriendshipRequest(models.Model):
 
     def reject(self):
         """ reject this friendship request """
-        self.rejected = datetime.datetime.now()
+        self.rejected = timezone.now()
         friendship_request_rejected.send(sender=self)
         self.save()
         bust_cache('requests', self.to_user.pk)
@@ -112,7 +115,7 @@ class FriendshipRequest(models.Model):
         return True
 
     def mark_viewed(self):
-        self.viewed = datetime.datetime.now()
+        self.viewed = timezone.now()
         friendship_request_viewed.send(sender=self)
         self.save()
         bust_cache('requests', self.to_user.pk)
@@ -141,7 +144,20 @@ class FriendshipManager(models.Manager):
 
         if requests is None:
             qs = FriendshipRequest.objects.select_related(depth=1).filter(
-                    to_user=user).all()
+                to_user=user).all()
+            requests = list(qs)
+            cache.set(key, requests)
+
+        return requests
+
+    def sent_requests(self, user):
+        """ Return a list of friendship requests from user """
+        key = cache_key('sent_requests', user.pk)
+        requests = cache.get(key)
+
+        if requests is None:
+            qs = FriendshipRequest.objects.select_related(depth=1).filter(
+                    from_user=user).all()
             requests = list(qs)
             cache.set(key, requests)
 
@@ -154,8 +170,8 @@ class FriendshipManager(models.Manager):
 
         if unread_requests is None:
             qs = FriendshipRequest.objects.select_related(depth=1).filter(
-                    to_user=user,
-                    viewed__isnull=True).all()
+                to_user=user,
+                viewed__isnull=True).all()
             unread_requests = list(qs)
             cache.set(key, unread_requests)
 
@@ -168,8 +184,8 @@ class FriendshipManager(models.Manager):
 
         if count is None:
             count = FriendshipRequest.objects.select_related(depth=1).filter(
-                    to_user=user,
-                    viewed__isnull=True).count()
+                to_user=user,
+                viewed__isnull=True).count()
             cache.set(key, count)
 
         return count
@@ -181,8 +197,8 @@ class FriendshipManager(models.Manager):
 
         if read_requests is None:
             qs = FriendshipRequest.objects.select_related(depth=1).filter(
-                    to_user=user,
-                    viewed__isnull=False).all()
+                to_user=user,
+                viewed__isnull=False).all()
             read_requests = list(qs)
             cache.set(key, read_requests)
 
@@ -195,8 +211,8 @@ class FriendshipManager(models.Manager):
 
         if rejected_requests is None:
             qs = FriendshipRequest.objects.select_related(depth=1).filter(
-                    to_user=user,
-                    rejected__isnull=False).all()
+                to_user=user,
+                rejected__isnull=False).all()
             rejected_requests = list(qs)
             cache.set(key, rejected_requests)
 
@@ -204,10 +220,14 @@ class FriendshipManager(models.Manager):
 
     def add_friend(self, from_user, to_user):
         """ Create a friendship request """
+        if from_user == to_user:
+            raise ValidationError("Users cannot be friends with themselves")
+
         request = FriendshipRequest.objects.create(
-                    from_user=from_user,
-                    to_user=to_user
-                )
+            from_user=from_user,
+            to_user=to_user
+        )
+
         bust_cache('requests', to_user.pk)
         friendship_request_created.send(sender=request)
 
@@ -217,15 +237,16 @@ class FriendshipManager(models.Manager):
         """ Destroy a friendship relationship """
         try:
             qs = Friend.objects.filter(
-                        Q(to_user=to_user, from_user=from_user) |
-                        Q(to_user=from_user, from_user=to_user)
-                    ).distinct().all()
+                Q(to_user=to_user, from_user=from_user) |
+                Q(to_user=from_user, from_user=to_user)
+            ).distinct().all()
+
             if qs:
                 friendship_removed.send(
-                        sender=qs[0],
-                        from_user=from_user,
-                        to_user=to_user
-                    )
+                    sender=qs[0],
+                    from_user=from_user,
+                    to_user=to_user
+                )
                 qs.delete()
                 bust_cache('friends', to_user.pk)
                 bust_cache('friends', from_user.pk)
@@ -255,7 +276,7 @@ class Friend(models.Model):
     """ Model to represent Friendships """
     to_user = models.ForeignKey(AUTH_USER_MODEL, related_name='friends')
     from_user = models.ForeignKey(AUTH_USER_MODEL, related_name='_unused_friend_relation')
-    created = models.DateTimeField(default=datetime.datetime.now)
+    created = models.DateTimeField(default=timezone.now)
 
     objects = FriendshipManager()
 
@@ -267,6 +288,12 @@ class Friend(models.Model):
     def __unicode__(self):
         return "User #%d is friends with #%d" % (self.to_user_id, self.from_user_id)
 
+    def save(self, *args, **kwargs):
+        # Ensure users can't be friends with themselves
+        if self.to_user == self.from_user:
+            raise ValidationError("Users cannot be friends with themselves.")
+        super(Friend, self).save(*args, **kwargs)
+
 
 class FollowingManager(models.Manager):
     """ Following manager """
@@ -297,6 +324,9 @@ class FollowingManager(models.Manager):
 
     def add_follower(self, follower, followee):
         """ Create 'follower' follows 'followee' relationship """
+        if follower == followee:
+            raise ValidationError("Users cannot follow themselves")
+
         relation = Follow.objects.create(follower=follower, followee=followee)
 
         follower_created.send(sender=self, follower=follower)
@@ -341,7 +371,7 @@ class Follow(models.Model):
     """ Model to represent Following relationships """
     follower = models.ForeignKey(AUTH_USER_MODEL, related_name='following')
     followee = models.ForeignKey(AUTH_USER_MODEL, related_name='followers')
-    created = models.DateTimeField(default=datetime.datetime.now)
+    created = models.DateTimeField(default=timezone.now)
 
     objects = FollowingManager()
 
@@ -352,3 +382,9 @@ class Follow(models.Model):
 
     def __unicode__(self):
         return "User #%d follows #%d" % (self.follower_id, self.followee_id)
+
+    def save(self, *args, **kwargs):
+        # Ensure users can't be friends with themselves
+        if self.follower == self.followee:
+            raise ValidationError("Users cannot follow themselves.")
+        super(Follow, self).save(*args, **kwargs)
